from __future__ import annotations
from enum import Enum
from math import floor


## Every size number of memory usage are in CacheLine 64 bytes
class MemoryLevelSize(Enum):
    L1 = float(32 * 1024 / 64)
    L2 = float(1024 * 1024 / 64)
    L3 = float(8448 * 1024 / 64)
    RAM = float(31 * 1024**3 / 64)


class MemoryLevel(Enum):
    L1 = 0
    L2 = 1
    L3 = 2
    RAM = 3

class Dimension(Enum):
    FIRST_DIMENSION = 0
    SECOND_DIMENSION = 1

class MemoryLevelAccess:

    def __init__(self, mem_level: MemoryLevel) -> None:
        self.mem_level: MemoryLevel = mem_level

    @property
    def read_cost(self) -> int:
        if self.mem_level == MemoryLevel.L1:
            cost = 3
        elif self.mem_level == MemoryLevel.L2:
            cost = 10 
        elif self.mem_level == MemoryLevel.L3:
            cost = 60
        else:
            assert self.mem_level == MemoryLevel.RAM
            cost = 100
        return  cost

    @property
    def write_cost(self) -> int:
        return 2* self.read_cost

class Access:

    FIRST_ARRAY_INDEX = 1

    def __init__(self, array_name: str, i: int, j: int, is_read_access: bool, cache_level_accessed: MemoryLevelAccess):
        self.array_name = array_name
        ##i is FIRST DIMENSION
        self.i = i
        ##j is SECOND DIMENSION
        self.j = j
        self.is_read_access = is_read_access
        self.cache_level_accessed = cache_level_accessed
        ## NOTE: 64 bytes per cache line, 4 bytes (32bits) per Single Precision Float
        # thus 16 Real per cache line
        self.nb_access_per_cache_line = 16

    def in_same_cache_line(self, previous_access: Access, contiguous_dimension: Dimension):
        """Assuming array is alined in memory among contiguous dimension: Access 0 in this dimension
            is the start of a cache line
        """
        def _in_same_cache_line(
            current_contiguous_access_index,
            previous_contiguous_access_index,
            current_noncontiguous_access_index,
            previous_noncontiguous_access_index,
        ) -> bool:
            if current_noncontiguous_access_index != previous_noncontiguous_access_index:
                ## not same indice on non-contiguous dimension
                return False
            else:
                current_cache_line = floor(
                    (current_contiguous_access_index - Access.FIRST_ARRAY_INDEX) / self.nb_access_per_cache_line
                )
                previous_cache_line = floor(
                    (previous_contiguous_access_index - Access.FIRST_ARRAY_INDEX) / self.nb_access_per_cache_line
                )
                if current_cache_line == previous_cache_line:
                    return True
                else:
                    return False
        if self.array_name.lower() != previous_access.array_name.lower():
            return False
        if contiguous_dimension == Dimension.FIRST_DIMENSION:
            return _in_same_cache_line(self.i, previous_access.i, self.j, previous_access.j)
        else:
            assert contiguous_dimension == Dimension.SECOND_DIMENSION
            return _in_same_cache_line(self.j, previous_access.j, self.i, previous_access.i)

    @property
    def static_cost(self):
        return 1 if self.is_read_access else 2

    @property
    def dynamic_cost(self):
        cache_access_cost = self.cache_level_accessed.read_cost  if self.is_read_access else self.cache_level_accessed.write_cost
        return self.static_cost * cache_access_cost

class PerfModelMain:

    def __init__(self, array_names: list[str]) -> None:
        self.contiguous_dimension = Dimension.FIRST_DIMENSION
        self.current_memory_footprint = {}
        for array_name in array_names:
            self.current_memory_footprint[array_name] = 0.0

    @property
    def current_total_memory_footprint(self):
        return sum(list(self.current_memory_footprint.values()))

    def access_cost(self, current_access: Access, last_access) -> float:
        "Returns the cost of current_access knowing last_access"
        if current_access.in_same_cache_line(last_access, self.contiguous_dimension):
            return 0.0
        else:
            self.current_memory_footprint[current_access.array_name] += 1
            return current_access.dynamic_cost

    @property
    def l1_exceeded(self):
        for memory_footprint in self.current_memory_footprint.values():
            if memory_footprint > MemoryLevelSize.L1.value:
                return True
        return False

    @property
    def l2_exceeded(self):
        for memory_footprint in self.current_memory_footprint.values():
            if memory_footprint > MemoryLevelSize.L2.value:
                return True
        return False

    @property
    def l3_exceeded(self):
        for memory_footprint in self.current_memory_footprint.values():
            if memory_footprint > MemoryLevelSize.L3.value:
                return True
        return False


def run():
    from perf_model.plot import Plot
    N = 1000
    M = 1000
    perf_model = PerfModelMain(["A", "B"])
    last_access_a = Access("A", 1, 1, True, MemoryLevelAccess(MemoryLevel.L1))
    last_access_b = Access("B", 1, 1, True, MemoryLevelAccess(MemoryLevel.L1))
    last_access_b_1 = Access("B", 1, 1, True, MemoryLevelAccess(MemoryLevel.L1))
    x_ticks: list[str] = []
    y_footprint = []
    y_footprint_a = []
    y_footprint_b = []
    xticklabels = []
    modn = 100000
    keep_list = [1, 50000]
    l1_passed = False
    l2_passed = False
    l3_passed = False
    for j in range(1, N):
        for i in range(1, M):
            current_x = str((i, j))
            if (i + M * (j - 1)) % modn in keep_list:
                x_ticks.append(current_x)
                xticklabels.append(current_x)
            # else:
            #    xticklabels.append("")
            current_access_a = Access("A", i, j, False, MemoryLevelAccess(MemoryLevel.L1))
            perf_model.access_cost(current_access_a, last_access_a)
            last_access_a = current_access_a

            current_access_a = Access("A", i - 1, j, True, MemoryLevelAccess(MemoryLevel.L1))
            perf_model.access_cost(current_access_a, last_access_a)
            last_access_a = current_access_a

            current_access_b = Access("B", i, j, True, MemoryLevelAccess(MemoryLevel.L1))
            perf_model.access_cost(current_access_b, last_access_b)
            last_access_b = current_access_b

            current_access_b_1 = Access("B", i, j + 1, True, MemoryLevelAccess(MemoryLevel.L1))
            perf_model.access_cost(current_access_b_1, last_access_b_1)
            last_access_b_1 = current_access_b_1

            if (i + M * (j - 1)) % modn in keep_list:
                y_footprint_a.append(perf_model.current_memory_footprint["A"])
                y_footprint_b.append(perf_model.current_memory_footprint["B"])
                y_footprint.append(perf_model.current_total_memory_footprint)
            if not l1_passed and perf_model.l1_exceeded:
                print(f"Exceed L1 cache, tile should be: I={M},J={j-1}")
                l1_passed = True
            if not l2_passed and perf_model.l2_exceeded:
                print(f"Exceed L2 cache, tile should be: I={M},J={j-1}")
                l2_passed = True
            if not l3_passed and perf_model.l3_exceeded:
                print(f"Exceed L3 cache, tile should be: I={M},J={j-1}")
                l3_passed = True

    assert len(x_ticks) == len(y_footprint_a)
    assert len(x_ticks) == len(y_footprint_b)
    assert len(x_ticks) == len(y_footprint)
    assert len(x_ticks) == len(xticklabels)
    plot = Plot(x_ticks, xticklabels)
    title = (
        "A and B of size (700,800) L3tile should be 700,744, Memory footprint of Stencil access A(i,j) = A(i,j) + A(i-1,j) +"
        " B(i,j) + B(i,j+1)"
    )
    y_l1cache_size = [MemoryLevelSize.L1.value] * len(x_ticks)
    y_l2cache_size = [MemoryLevelSize.L2.value] * len(x_ticks)
    y_l3cache_size = [MemoryLevelSize.L3.value] * len(x_ticks)
    plot.run_line("L1 capacity", y_l1cache_size, color="green")
    plot.run_line("L2 capacity", y_l2cache_size, color="blue")
    plot.run_line("L3 capacity", y_l3cache_size, color="red")
    plot.run_scatter("A footprint", y_footprint_a, marker="o", color="blue", size=20)
    plot.run_scatter("B footprint", y_footprint_b, marker="x", color="grey", size=20)
    plot.run_scatter("Total footprint", y_footprint, marker=".", color="black", size=20)
    plot.gen_figure(title)
